from typing import Any

from model.m3 import Class, Property


class PropertyMapping:
    def __init__(self, property: Property, target: Any):
        self.property = property
        self.target = target

class ClassMapping:
    def __init__(self, clazz: Class, property_mappings: list[PropertyMapping]):
        self.clazz = clazz
        self.property_mappings = property_mappings

class Mapping:
    def __init__(self, name: str, mappings: list[ClassMapping]):
        self.name = name
        self.mappings = mappings


import os

from model.m3 import PrimitiveType, Property
from jinja2 import Environment, PackageLoader

from model.mapping import Mapping


def is_primitive(prop: Property) -> bool:
    return isinstance(prop.type, PrimitiveType)


def generate(mapping:Mapping, output_directory):
    environment = Environment(loader=PackageLoader("datafinder_generator"), trim_blocks=True, lstrip_blocks=True)
    template = environment.get_template("finder_template.txt")

    for rcm in mapping.mappings:
        filename = f"{rcm.clazz.name.lower()}_finder.py"
        filepath = os.path.join(output_directory, filename)
        content = template.render(rcm=rcm,is_primitive=is_primitive)
        with open(filepath, mode="w", encoding="utf-8") as message:
            message.write(content)
            print(f"... wrote {filename}")





import datetime

from datafinder.attribute import Attribute


class TableAlias:
    def __init__(self, table: str, alias: str):
        self.table = table
        self.alias = alias


class ColumnAlias:
    def __init__(self, column_name: str, table_alias: TableAlias):
        self.column_name = column_name
        self.table_alias = table_alias


class Join:
    def __init__(self, source: ColumnAlias, target: ColumnAlias):
        self.source = source
        self.target = target


class QueryEngine:
    _select: list[ColumnAlias]
    _from: set[TableAlias]
    _where: list[str]
    _join: list[Join]
    __table_alias_incr: int

    def __init__(self):
        self._where = []
        self._select = []
        self._from = set()
        self._join = []
        self.__table_alias_incr = 0
        self.__table_aliases_by_table = {}

    def select(self, cols: list[Attribute]):
        for col in cols:
            table = col.owner()
            ta = self.__table_alias_for_table(table)
            parent: JoinOperation = col.parent()
            if parent is not None:
                left = parent.left
                sc = ColumnAlias(left.column_name(), self.__table_alias_for_table(left.owner()))
                right = parent.right
                tc = ColumnAlias(right.column_name(), self.__table_alias_for_table(right.owner()))
                self._join.append(Join(sc, tc))
            else:
                self._from.add(ta)
            ca = ColumnAlias(col.column_name(), ta)
            self._select.append(ca)

    def __table_alias_for_table(self, table: str) -> TableAlias:
        ta = None
        if table in self.__table_aliases_by_table:
            ta = self.__table_aliases_by_table[table]
        else:
            ta = TableAlias(table, "t" + str(self.__table_alias_incr))
            self.__table_alias_incr = self.__table_alias_incr + 1
            self.__table_aliases_by_table[table] = ta
        return ta

    def append_where_binary_clause(self, op: str):
        self._where.append(op)

    def append_where_clause(self, attr: Attribute, op: str, value: str):
        ta = self.__table_alias_for_table(attr.owner())
        self._where.append(ta.alias + '.' + attr.column_name() + ' ' + op + ' ' + value)

    def build_query_string(self) -> str:
        joins = map(lambda j: ' LEFT OUTER JOIN ' + j.target.table_alias.table + ' AS ' + j.target.table_alias.alias +
                              ' ON ' + j.source.table_alias.alias + '.' + j.source.column_name + ' = ' +
                              j.target.table_alias.alias + '.' + j.target.column_name, self._join)
        return 'SELECT ' + ','.join(map(lambda ca: ca.table_alias.alias + '.' + ca.column_name, self._select)) \
            + ' FROM ' + ','.join(map(lambda ta: ta.table + ' AS ' + ta.alias, self._from)) \
            + ''.join(joins) \
            + self.__build_where()

    def __build_where(self) -> str:
        if len(self._where) > 0:
            return ' WHERE ' + ''.join(self._where)
        else:
            return ''

    def where_clauses(self):
        return self._where

    def start_and(self):
        pass

    def end_and(self):
        pass


# Interface
class Operation:

    def generate_query(self, query: QueryEngine):
        pass


class NoOperation(Operation):
    def __init__(self):
        pass


class SelectOperation(Operation):
    def __init__(self, display: list[Attribute], table: str, filter: Operation):
        self.__display = display
        self.__table = table
        self.__filter = filter

    def generate_query(self, qe: QueryEngine):
        qe.select(self.__display)
        self.__filter.generate_query(qe)


class AndOperation(Operation):
    __left: Operation
    __right: Operation

    def __init__(self, lhs: Operation, rhs: Operation):
        self.__left = lhs
        self.__right = rhs

    def generate_query(self, query: QueryEngine):
        query.start_and()
        self.__left.generate_query(query)
        query.append_where_binary_clause(" and ")
        self.__right.generate_query(query)
        query.end_and()


class BusinessTemporalOperation(Operation):
    # TODO - which date format should we use
    __business_date_from_inclusive: datetime.date
    __business_date_to_inclusive: datetime.date


class BaseOperation(Operation):

    def and_op(self, rhs: Operation):
        return AndOperation(self, rhs)


class JoinOperation(Operation):
    left: Attribute
    right: Attribute

    def __init__(self, lhs: Attribute, rhs: Attribute):
        self.left = lhs
        self.right = rhs


def select_sql_to_string(columns: list[Attribute], table: str, op: Operation) -> str:
    qe = QueryEngine()
    select = SelectOperation(columns, table, op)
    select.generate_query(qe)
    return qe.build_query_string()




from .operation import BaseOperation, QueryEngine
from .attribute import Attribute


class EqOperation(BaseOperation):
    __attribute: Attribute

    def __init__(self, attrib: Attribute):
        self.__attribute = attrib

    def attribute(self) -> Attribute:
        return self.__attribute

    def column_type(self) -> str:
        return self.__attribute.column_type()

    def column_name(self) -> str:
        return self.__attribute.column_name()

    def prepare_value(self) -> str:
        pass


class PrimitiveEqOperation(EqOperation):
    __value: []

    def __init__(self, attrib: Attribute, value):
        super().__init__(attrib)
        self.__value = value

    def generate_query(self, query: QueryEngine):
        query.append_where_clause(self.attribute(), '=', self.prepare_value())

    def prepare_value(self) -> str:
        return str(self.__value)


class StringEqOperation(EqOperation):
    __value: str

    def __init__(self, attrib: Attribute, value: str):
        super().__init__(attrib)
        self.__value = value

    def generate_query(self, query: QueryEngine):
        query.append_where_clause(self.attribute(), 'LIKE', self.prepare_value())

    def prepare_value(self) -> str:
        #TODO - String escaper
        if self.column_type() == 'kx_symbol':
            return "`" + self.__value
        else:
            return "'" + self.__value + "'"


class GreaterThanOperation(BaseOperation):
    __attribute: Attribute

    def column_type(self) -> str:
        return self.__attribute.column_type()

    def __init__(self, attrib: Attribute):
        self.__attribute = attrib

    def generate_query(self, query: QueryEngine):
        query.append_where_clause(self.__attribute, '>', self.prepare_value())

    def prepare_value(self) -> str:
        pass


class PrimitiveGreaterThanOperation(GreaterThanOperation):
    __value: []

    def __init__(self, attrib: Attribute, value):
        super().__init__(attrib)
        self.__value = value

    def prepare_value(self) -> str:
        return str(self.__value)



from model.m3 import Property, Class
from model.mapping import ClassMapping, PropertyMapping


class RelationalElement:
    def __init(self):
        pass


class Column(RelationalElement):
    def __init__(self, name: str, type: str):
        self.name = name
        self.type = type
        self.table = None


class Table:
    def __init__(self, name: str, columns: list[Column]):
        self.name = name
        self.columns = columns
        for col in columns:
            col.table = self


class Join(RelationalElement):
    def __init__(self, lhs: Column, rhs: Column):
        self.lhs = lhs
        self.rhs = rhs


class RelationalPropertyMapping(PropertyMapping):
    def __init__(self, property: Property, target: RelationalElement):
        super().__init__(property, target)


class RelationalClassMapping(ClassMapping):
    def __init__(self, clazz: Class, property_mappings: list[RelationalPropertyMapping]):
        super().__init__(clazz, property_mappings)






from typing import Any


class Attribute:
    __name: str
    __column_db_type: str
    __owner: str
    __parent: Any

    def __init__(self, name: str, column_db_type: str, owner:str, parent=None):
        self.__name = name
        self.__column_db_type = column_db_type
        self.__owner = owner
        self.__parent = parent

    def column_name(self) -> str:
        return self.__name

    def column_type(self) -> str:
        return self.__column_db_type

    def owner(self) -> str:
        return self.__owner

    def parent(self) -> Any:
        return self.__parent



from .operation import *
from .typed_operations import *


class StringAttribute(Attribute):

    def __init__(self, name: str, column_db_type: str, owner:str, parent=None):
        super().__init__(name, column_db_type, owner, parent)

    def eq(self, value: str) -> Operation:
        return StringEqOperation(self, value)

    def __eq__(self, value: str) -> Operation:
        return StringEqOperation(self, value)


class FloatAttribute(Attribute):

    def __init__(self, name: str, column_db_type: str, owner:str, parent=None):
        super().__init__(name, column_db_type, owner, parent)

    def eq(self, value: float) -> Operation:
        return PrimitiveEqOperation(self, value)

    def __eq__(self, value: float) -> Operation:
        return PrimitiveEqOperation(self, value)

    def __gt__(self, value: float) -> Operation:
        return PrimitiveGreaterThanOperation(self, value)

class IntegerAttribute(Attribute):

    def __init__(self, name: str, column_db_type: str, owner:str, parent=None):
        super().__init__(name, column_db_type, owner, parent)

    def eq(self, value: float) -> Operation:
        return PrimitiveEqOperation(self, value)

    def __eq__(self, value: float) -> Operation:
        return PrimitiveEqOperation(self, value)

    def __gt__(self, value: float) -> Operation:
        return PrimitiveGreaterThanOperation(self, value)



from datafinder import Operation, DataFrame, Attribute, select_sql_to_string

import ibis
import numpy as np
import pandas as pd

from datafinder import QueryRunnerBase


class IbisConnect(QueryRunnerBase):

    @staticmethod
    def select(columns: list[Attribute], table: str, op: Operation) -> DataFrame:
        conn = ibis.connect('duckdb://test.db')
        query = select_sql_to_string(columns, table, op)
        print(query)
        t = conn.table(table)
        #todo - can also do this with the dataframe API
        return IbisOutput(t.sql(query))


class IbisOutput(DataFrame):

    def __init__(self, t: ibis.Table):
        self.__table = t

    def to_numpy(self) -> np.array:
        return self.__table.__array__()

    def to_pandas(self) -> pd.DataFrame:
        return self.__table.to_pandas()



from datafinder import Operation, DataFrame, Attribute, select_sql_to_string, QueryRunnerBase

import duckdb
import numpy as np
import pandas as pd


class DuckDbConnect(QueryRunnerBase):

    @staticmethod
    def select(columns: list[Attribute], table: str, op: Operation) -> DataFrame:
        conn = duckdb.connect('test.db')
        query = select_sql_to_string(columns, table, op)
        print(query)
        # TODO this is inefficient, could convert straight to desired output - such as numpy, instead of list
        return DuckDbOutput(conn.sql(query).fetchall())


class DuckDbOutput(DataFrame):
    __table: list

    def __init__(self, t: list):
        self.__table = t

    def to_numpy(self) -> np.array:
        #TODO - this could be a better dtype
        return np.array(self.__table, dtype='object')

    def to_pandas(self) -> pd.DataFrame:
        #todo - this needs to be better, to ensure types and column names
        return pd.DataFrame(self.__table)



from datafinder import Attribute, Operation, DataFrame


#TODO revisit this, don't want this to be static per class as need to be able to switch them
class RegistryBase(type):
    REGISTRY = {}

    def __new__(cls, name, bases, attrs):
        # instantiate a new type corresponding to the type of class being defined
        # this is currently RegisterBase but in child classes will be the child class
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY[new_cls.__name__] = new_cls
        return new_cls

    @classmethod
    def get_registry(cls):
        return dict(cls.REGISTRY)

    @classmethod
    def register(cls, clazz):
        cls.REGISTRY[clazz.__name__] = clazz

    @classmethod
    def clear(cls):
        RegistryBase.REGISTRY = {}


class QueryRunnerBase(metaclass=RegistryBase):

    @staticmethod
    def select(columns: list[Attribute], table: str, op: Operation) -> DataFrame:
        pass

    @staticmethod
    def get_runner():
        for k in RegistryBase.REGISTRY.keys():
            if k != 'QueryRunnerBase':
                return RegistryBase.REGISTRY[k]
        raise Exception("No query runner registered")



class Package:
    def __init__(self, name: str):
        self.name = name


class PackagableElement:
    def __init__(self, package: Package):
        self.package = package


class Type:
    def __init(self):
        pass


class PrimitiveType(Type):
    def __init__(self, name: str):
        self.name = name


Integer = PrimitiveType("Integer")
String = PrimitiveType("String")
Float = PrimitiveType("Float")
DateTime = PrimitiveType("DateTime")
Date = PrimitiveType("Date")


class Property:
    def __init__(self, name: str, type: Type):
        self.name = name
        self.type = type


class Class(PackagableElement, Type):
    def __init__(self, name: str, properties: list[Property], package: Package):
        super().__init__(package)
        self.properties = {}
        self.name = name
        for prop in properties:
            self.properties[prop.name] = prop

    def property(self, name:str) -> Property:
        return self.properties[name]


class Association(PackagableElement):
    def __init__(self, name: str, source: str, target: str, package: Package):
        super().__init__(package)
        self.name = name
        self.source = source
        self.target = target
